
import java.awt.*;

/**

Television.java - Copyright 2003-2008 by Michael Kohn 

Part of the Jatari project.  This module falls under the GPL license.

1 2 1 2 The Naken Crew

Email: mike@mikekohn.net
  Web: http://www.mikekohn.net/

Horizontal Scan: 68 clocks + 160 clock = 228 clocks
Vertical Scan: 3 clocks + 37 clocks + 192 clocks + 30 clocks = 262 clocks
Total Resolution: 228x262
Visible Resolution: 160x192

Clock Speed: 3.58MHz  (CPU clock * 3)

*/

public class Television extends Frame
{
final int TV_WIDTH=228;
final int TV_HEIGHT=262;
public int posx=0;
public int posy=0;
Color[] colors;
Image tvImage=null;
Graphics tvgg=null;
float pixel_size;
float rasterx;
int width,height;
boolean okay=false;
int offsetY=0;
long startTime;
int frame_count;

  public Television(int width, int height)
  {
    super();

    this.setSize(width,height);
    this.setVisible(true);

    this.width=width;
    this.height=height;

    pixel_size=(float)width/160;
    rasterx=0;

    // startTime=System.currentTimeMillis();
  }

  public void init()
  {
  Insets ins;

    ins = this.getInsets();
    this.setSize(width+ins.left+ins.right,height+ins.top+ins.bottom);
    offsetY=ins.top;

    setupColorTable();

    // OMG
    while(tvImage==null)
    { tvImage=createImage(width,height); }
    // OMG again, wtf
    while(tvgg==null)
    { tvgg=tvImage.getGraphics(); }
    tvgg.setColor(colors[0]);
  }

  /** Define all colors. */

  private void setupColorTable()
  {
  int c,cx,y,r,i;
  int[][] yuv_table=
    {
      { 255, 128, 128 },     // Grey 0
      { 235,  54, 140 },     // Gold 1
      { 201,  81, 152 },     // Orange 2
      { 202,  97, 163 },     // Bright Orange 3
      { 201, 115, 164 },     // Pink 4
      { 199, 141, 154 },     // Purple 5
      { 195, 159, 159 },     // Purple-Blue 6
      { 188, 180, 252 },     // Blue 7
      { 144, 172, 113 },     // Blue 8
      { 195, 160, 105 },     // Light-Blue 9
      { 166, 153,  97 },     // Torquois 10 
      { 221, 122,  87 },     // Green-Blue 11
      { 223, 105,  99 },     // Green-Yellow 12
      { 226,  92, 109 },     // Green 13
      { 223,  90, 128 },     // Orange-Green 14
      { 180,  87, 147 }      // Light-Orange 15
    };

    colors=new Color[128];
    c=0;

    while(c<128)
    {
      if (c==0)
      { y=0; }
        else
      { y=8; }

      i=c/8;
      cx=yuv_table[i][0]/8;
      for (r=0; r<8; r++)
      {
        colors[c++]=yuv2Color(y,yuv_table[i][1],yuv_table[i][2]);
        y=y+cx;
      }
    }
  }

  /** Get the current raster beam X position (includes blanking area) */

  public int getX()
  {
    return posx;
  }

  /** Get the current raster beam Y position (includes blanking area) */

  public int getY()
  {
    return posy;
  }

  /** Apply 1 pixel clock.  Draw the current raster beam color in this area */

  public void clock()
  {
    if (posx>=68 && posy>40 && posy<232)
    {
      tvgg.drawLine((int)rasterx,posy-40,(int)(rasterx+pixel_size),posy-40);
      // tvgg.drawLine((int)rasterx,posy,(int)(rasterx+pixel_size),posy);
      rasterx=rasterx+pixel_size;
    }

    posx++;
    if (posx>=TV_WIDTH)
    {
      // Hsync!
      posx=0;
      rasterx=0;
      posy++;
    }
  }

  /** Apply n pixel clocks.  Draw the current raster beam color in this area */

  public void clock(int n)
  {
  int c,blank;

    while(n>0)
    {
      c=TV_WIDTH-posx;
      if (c>n) c=n;
      n=n-c;
      if (posx<68)
      {
        blank=68-posx;
        if (blank<=c)
        {
          posx=posx+c;
          continue;
        }
        c=c-blank;
        posx=68;
      }

      if (posx>=68 && posy>40 && posy<232)
      {
        tvgg.drawLine((int)rasterx,posy-40,(int)(rasterx+(c*pixel_size)),posy-40);
      }

      posx=posx+c;
      if (posx>=TV_WIDTH)
      {
        // Hsync!
        posx=0;
        rasterx=0;
        posy++;
      }
    }
  }

  /** Apply vsync pulse to TV.  Usually this lasts 3 scan lines but
     we'll assume all is okay from a single pulse */

  public void vsync()
  {
    repaint();
    posx=0;
    posy=0;

    frame_count++;

    if (System.currentTimeMillis()-startTime>=1000)
    {
      System.out.println("fps="+(float)(frame_count)/((float)(System.currentTimeMillis()-startTime)/1000));
      startTime=System.currentTimeMillis();
      frame_count=0;
    }
  }

  /** Change the raster beam color to a preset Atari color */

  public void setColor(int c)
  {
    tvgg.setColor(colors[c&0x7f]);
  }

  /** Change the raster beam color to a YUV color */

  public void setColor(int y, int u, int v)
  {
    tvgg.setColor(yuv2Color(y,u,v));
  }

  /** Paint the screen. Not needed to be called by other classes. */

  public void update(Graphics gg)
  {
    gg.drawImage(tvImage,0,offsetY,null);
  }

  public void paint(Graphics gg)
  {
    gg.drawImage(tvImage,0,offsetY,null);
  }

  /** Create a new Java color from y,u,v */

  Color yuv2Color(int y, int u, int v)
  {
  double r,g,b;

    u=u-128;
    v=v-128;

    r=(double)y+(double)(1.4075 * ((double)v));
    g=(double)y-(double)(0.3455 * ((double)u)) - (double)(0.7169*((double)v));
    b=(double)y+(double)(1.7790 * ((double)u));

    if (r<0) r=0;
    if (g<0) g=0;
    if (b<0) b=0;

    if (r>255) r=255;
    if (g>255) g=255;
    if (b>255) b=255;

    return new Color((int)r,(int)g,(int)b);
  }
}


